/*----- PROTECTED REGION ID(UniversalTestClass.cpp) ENABLED START -----*/
//=============================================================================
//
// file :        UniversalTestClass.cpp
//
// description : C++ source for the UniversalTestClass.
//               A singleton class derived from DeviceClass.
//               It implements the command and attribute list
//               and all properties and methods required
//               by the UniversalTest once per process.
//
// project :     UniversalTest
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author:  $
// Copyright (C): 2015-2019
//                European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                France
//
// $Revision:  $
// $Date:  $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include "UniversalTestClass.h"

/*----- PROTECTED REGION END -----*/	//	UniversalTestClass.cpp

//-------------------------------------------------------------------
/**
 *	Create UniversalTestClass singleton and
 *	return it in a C function for Python usage
 */
//-------------------------------------------------------------------
extern "C" {
#ifdef _TG_WINDOWS_

__declspec(dllexport)

#endif

	Tango::DeviceClass *_create_UniversalTest_class(const char *name) {
		return UniversalTest_ns::UniversalTestClass::init(name);
	}
}

namespace UniversalTest_ns
{
//===================================================================
//	Initialize pointer for singleton pattern
//===================================================================
UniversalTestClass *UniversalTestClass::_instance = NULL;

//===================================================================
//	Class constants
//===================================================================
//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::UniversalTestClass(std::string &s)
 * description : 	constructor for the UniversalTestClass
 *
 * @param s	The class name
 */
//--------------------------------------------------------
UniversalTestClass::UniversalTestClass(std::string &s):Tango::DeviceClass(s)
{
	TANGO_LOG_INFO << "Entering UniversalTestClass constructor" << std::endl;
	set_default_property();
	write_class_property();

	/*----- PROTECTED REGION ID(UniversalTestClass::constructor) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::constructor

	TANGO_LOG_INFO << "Leaving UniversalTestClass constructor" << std::endl;
}

//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::~UniversalTestClass()
 * description : 	destructor for the UniversalTestClass
 */
//--------------------------------------------------------
UniversalTestClass::~UniversalTestClass()
{
	/*----- PROTECTED REGION ID(UniversalTestClass::destructor) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::destructor

	_instance = NULL;
}


//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::init
 * description : 	Create the object if not already done.
 *                  Otherwise, just return a pointer to the object
 *
 * @param	name	The class name
 */
//--------------------------------------------------------
UniversalTestClass *UniversalTestClass::init(const char *name)
{
	if (_instance == NULL)
	{
		try
		{
			std::string s(name);
			_instance = new UniversalTestClass(s);
		}
		catch (std::bad_alloc &)
		{
			throw;
		}
	}
	return _instance;
}

//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::instance
 * description : 	Check if object already created,
 *                  and return a pointer to the object
 */
//--------------------------------------------------------
UniversalTestClass *UniversalTestClass::instance()
{
	if (_instance == NULL)
	{
		std::cerr << "Class is not initialized !!" << std::endl;
		exit(-1);
	}
	return _instance;
}



//===================================================================
//	Command execution method calls
//===================================================================

//===================================================================
//	Properties management
//===================================================================
//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::get_class_property()
 *	Description: Get the class property for specified name.
 */
//--------------------------------------------------------
Tango::DbDatum UniversalTestClass::get_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_prop.size() ; i++)
		if (cl_prop[i].name == prop_name)
			return cl_prop[i];
	//	if not found, returns  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::get_default_device_property()
 *	Description: Return the default value for device property.
 */
//--------------------------------------------------------
Tango::DbDatum UniversalTestClass::get_default_device_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<dev_def_prop.size() ; i++)
		if (dev_def_prop[i].name == prop_name)
			return dev_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}

//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::get_default_class_property()
 *	Description: Return the default value for class property.
 */
//--------------------------------------------------------
Tango::DbDatum UniversalTestClass::get_default_class_property(std::string &prop_name)
{
	for (unsigned int i=0 ; i<cl_def_prop.size() ; i++)
		if (cl_def_prop[i].name == prop_name)
			return cl_def_prop[i];
	//	if not found, return  an empty DbDatum
	return Tango::DbDatum(prop_name);
}


//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::set_default_property()
 *	Description: Set default property (class and device) for wizard.
 *                For each property, add to wizard property name and description.
 *                If default value has been set, add it to wizard property and
 *                store it in a DbDatum.
 */
//--------------------------------------------------------
void UniversalTestClass::set_default_property()
{
	std::string	prop_name;
	std::string	prop_desc;
	std::string	prop_def;
	std::vector<std::string>	vect_data;

	//	Set Default Class Properties

	//	Set Default device Properties
}

//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::write_class_property()
 *	Description: Set class description fields as property in database
 */
//--------------------------------------------------------
void UniversalTestClass::write_class_property()
{
	//	First time, check if database used
	if (Tango::Util::_UseDb == false)
		return;

	Tango::DbData	data;
	std::string	classname = get_name();
	std::string	header;

	//	Put title
	Tango::DbDatum	title("ProjectTitle");
	std::string	str_title("UniversalTest");
	title << str_title;
	data.push_back(title);

	//	Put Description
	Tango::DbDatum	description("Description");
	std::vector<std::string>	str_desc;
	str_desc.push_back("The goal of this device server is to be able to provide attributes of every possible Tango type.");
	str_desc.push_back("It is possible to change dynamically the Quality Factor of a given attribute.");
	str_desc.push_back("It is possible to make an attribute throw exceptions.");
	str_desc.push_back("It is possible to change the behaviour of the attributes (increment at each read, read value always set to the set value, ");
	str_desc.push_back("read value fluctuating around the set value,...)");
	str_desc.push_back("Some dynamic expert attributes will be created to be able to control the behaviour of the other attributes.");
	str_desc.push_back("The attributes named xxx_quality will allow to change the quality factor of the attribute named xxx.");
	description << str_desc;
	data.push_back(description);

	//  Put inheritance
	Tango::DbDatum	inher_datum("InheritedFrom");
	std::vector<std::string> inheritance;
	inheritance.push_back("TANGO_BASE_CLASS");
	inher_datum << inheritance;
	data.push_back(inher_datum);

	//	Call database and and values
	get_db_class()->put_property(data);
}

//===================================================================
//	Factory methods
//===================================================================

//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::device_factory()
 *	Description: Create the device object(s)
 *                and store them in the device list
 */
//--------------------------------------------------------
void UniversalTestClass::device_factory(const Tango::DevVarStringArray *devlist_ptr)
{
	/*----- PROTECTED REGION ID(UniversalTestClass::device_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::device_factory_before

	//	Create devices and add it into the device list
	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		TANGO_LOG_DEBUG << "Device name : " << (*devlist_ptr)[i].in() << std::endl;
		device_list.push_back(new UniversalTest(this, (*devlist_ptr)[i]));
	}

	//	Manage dynamic attributes if any
	erase_dynamic_attributes(devlist_ptr, get_class_attr()->get_attr_list());

	//	Export devices to the outside world
	for (unsigned long i=1 ; i<=devlist_ptr->length() ; i++)
	{
		//	Add dynamic attributes if any
		UniversalTest *dev = static_cast<UniversalTest *>(device_list[device_list.size()-i]);
		dev->add_dynamic_attributes();

		//	Check before if database used.
		if ((Tango::Util::_UseDb == true) && (Tango::Util::_FileDb == false))
			export_device(dev);
		else
			export_device(dev, dev->get_name().c_str());
	}

	/*----- PROTECTED REGION ID(UniversalTestClass::device_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::device_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::attribute_factory()
 *	Description: Create the attribute object(s)
 *                and store them in the attribute list
 */
//--------------------------------------------------------
void UniversalTestClass::attribute_factory(std::vector<Tango::Attr *> &att_list)
{
	/*----- PROTECTED REGION ID(UniversalTestClass::attribute_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::attribute_factory_before
	//	Attribute : setState
	setStateAttrib	*setstate = new setStateAttrib();
	Tango::UserDefaultAttrProp	setstate_prop;
	setstate_prop.set_description("This represents the state of the device. \nChanging this attribute set_value will invoke the set_state() method with the specified state value.");
	setstate_prop.set_label("setState");
	//	unit	not set for setState
	//	standard_unit	not set for setState
	//	display_unit	not set for setState
	//	format	not set for setState
	//	max_value	not set for setState
	//	min_value	not set for setState
	//	max_alarm	not set for setState
	//	min_alarm	not set for setState
	//	max_warning	not set for setState
	//	min_warning	not set for setState
	//	delta_t	not set for setState
	//	delta_val	not set for setState
	setstate->set_default_properties(setstate_prop);
	//	Not Polled
	setstate->set_disp_level(Tango::EXPERT);
	//	Not Memorized
	att_list.push_back(setstate);


	//	Create a list of static attributes
	create_static_attribute_list(get_class_attr()->get_attr_list());
	/*----- PROTECTED REGION ID(UniversalTestClass::attribute_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::attribute_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::pipe_factory()
 *	Description: Create the pipe object(s)
 *                and store them in the pipe list
 */
//--------------------------------------------------------
void UniversalTestClass::pipe_factory()
{
	/*----- PROTECTED REGION ID(UniversalTestClass::pipe_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::pipe_factory_before
	/*----- PROTECTED REGION ID(UniversalTestClass::pipe_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::pipe_factory_after
}
//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::command_factory()
 *	Description: Create the command object(s)
 *                and store them in the command list
 */
//--------------------------------------------------------
void UniversalTestClass::command_factory()
{
	/*----- PROTECTED REGION ID(UniversalTestClass::command_factory_before) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::command_factory_before


	/*----- PROTECTED REGION ID(UniversalTestClass::command_factory_after) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::command_factory_after
}

//===================================================================
//	Dynamic attributes related methods
//===================================================================

//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::create_static_attribute_list
 * description : 	Create the a list of static attributes
 *
 * @param	att_list	the created attribute list
 */
//--------------------------------------------------------
void UniversalTestClass::create_static_attribute_list(std::vector<Tango::Attr *> &att_list)
{
	for (unsigned long i=0 ; i<att_list.size() ; i++)
	{
		std::string att_name(att_list[i]->get_name());
		transform(att_name.begin(), att_name.end(), att_name.begin(), ::tolower);
		defaultAttList.push_back(att_name);
	}

	TANGO_LOG_INFO << defaultAttList.size() << " attributes in default list" << std::endl;

	/*----- PROTECTED REGION ID(UniversalTestClass::create_static_att_list) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::create_static_att_list
}


//--------------------------------------------------------
/**
 * method : 		UniversalTestClass::erase_dynamic_attributes
 * description : 	delete the dynamic attributes if any.
 *
 * @param	devlist_ptr	the device list pointer
 * @param	list of all attributes
 */
//--------------------------------------------------------
void UniversalTestClass::erase_dynamic_attributes(const Tango::DevVarStringArray *devlist_ptr, std::vector<Tango::Attr *> &att_list)
{
	Tango::Util *tg = Tango::Util::instance();

	for (unsigned long i=0 ; i<devlist_ptr->length() ; i++)
	{
		Tango::DeviceImpl *dev_impl = tg->get_device_by_name(((std::string)(*devlist_ptr)[i]).c_str());
		UniversalTest *dev = static_cast<UniversalTest *> (dev_impl);

		std::vector<Tango::Attribute *> &dev_att_list = dev->get_device_attr()->get_attribute_list();
		std::vector<Tango::Attribute *>::iterator ite_att;
		for (ite_att=dev_att_list.begin() ; ite_att != dev_att_list.end() ; ++ite_att)
		{
			std::string att_name((*ite_att)->get_name_lower());
			if ((att_name == "state") || (att_name == "status"))
				continue;
			std::vector<std::string>::iterator ite_str = find(defaultAttList.begin(), defaultAttList.end(), att_name);
			if (ite_str == defaultAttList.end())
			{
				TANGO_LOG_INFO << att_name << " is a UNWANTED dynamic attribute for device " << (*devlist_ptr)[i] << std::endl;
				Tango::Attribute &att = dev->get_device_attr()->get_attr_by_name(att_name.c_str());
				dev->remove_attribute(att_list[att.get_attr_idx()], true, false);
				--ite_att;
			}
		}
	}
	/*----- PROTECTED REGION ID(UniversalTestClass::erase_dynamic_attributes) ENABLED START -----*/
	
	/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::erase_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Method     : UniversalTestClass::get_attr_object_by_name()
 *	Description: returns Tango::Attr * object found by name
 */
//--------------------------------------------------------
Tango::Attr *UniversalTestClass::get_attr_object_by_name(std::vector<Tango::Attr *> &att_list, std::string attname)
{
	std::vector<Tango::Attr *>::iterator it;
	for (it=att_list.begin() ; it<att_list.end() ; ++it)
		if ((*it)->get_name()==attname)
			return (*it);
	//	Attr does not exist
	return NULL;
}


/*----- PROTECTED REGION ID(UniversalTestClass::Additional Methods) ENABLED START -----*/

/*----- PROTECTED REGION END -----*/	//	UniversalTestClass::Additional Methods
} //	namespace
